#!/bin/sh
# Check retrieving a non existent keyid fails

set -e

cd t
../onak -b -c test.conf add < ../keys/noodles.key
if ! ../onak -c test.conf get 0x12345678 2>&1 | \
	grep -q 'Key not found'; then
	echo "* Did not correctly error on non-existent key"
	exit 1
fi

exit 0
